"""The serve module provides a simple way to serve a model using Jina."""
import json
import logging

import jina
from jina import Document, DocumentArray
from jina import Gateway as BaseGateway
from jina.serve.runtimes.servers.composite import CompositeServer
from pydantic import BaseModel, Field


class GenerateRequest(BaseModel):
    prompt: str = Field(..., description='The prompt to generate from.')

    # session id
    id: str = Field(description='The session id of the generation.', default=None)

    # generation parameters
    num_beams: int = Field(description='The number of beams to use.', default=None)
    max_tokens: int = Field(
        description='The maximum length of the generated text.', default=None
    )
    temperature: float = Field(
        description='The temperature of the generation.', default=None
    )
    top_k: int = Field(description='The top k of the generation.', default=None)
    top_p: float = Field(description='The top p of the generation.', default=None)
    repetition_penalty: float = Field(
        description='The repetition penalty of the generation.', default=None
    )
    logprobs: int = Field(
        description='Include the log probabilities on the logprobs '
        'most likely tokens, as well the chosen tokens',
        default=None,
    )
    echo: bool = Field(
        description='Echo back the prompt in the completion.', default=None
    )
    stop: str = Field(description='Stop sequence generation on token.', default=None)
    do_sample: bool = Field(
        description='Whether to sample from the generation.', default=None
    )
    n: int = Field(description='The number of sequences to return.', default=None)

    class Config:
        allow_population_by_field_name = True
        arbitrary_types_allowed = True

        schema_extra = {
            'example': {
                'prompt': 'Hello, my name is',
                'id': '18d92585-7b66-4b7c-b818-71287c122c57',
                'num_beams': 5,
                'max_tokens': 50,
                'temperature': 0.7,
                'top_k': 50,
                'top_p': 0.95,
                'repetition_penalty': 1.0,
                'echo': False,
                'stop': '\n',
                'do_sample': True,
                'logprobs': None,
                'n': 3,
            }
        }


class Gateway(BaseGateway, CompositeServer):
    """A simple Jina Gateway that can be used to serve a generation model."""

    def __init__(self, **kwargs):
        """Initialize a new Gateway."""

        super().__init__(**kwargs)

        from fastapi import Body, status
        from fastapi.responses import JSONResponse

        def _update_key(parameters):
            key_maps = {
                'max_tokens': 'max_new_tokens',
                'n': 'num_return_sequences',
                'stop': 'stop_str',
            }
            for openai_key, hf_key in key_maps.items():
                if openai_key in parameters:
                    parameters[hf_key] = parameters.pop(openai_key)
            return parameters

        def _extend_rest_function(app):
            @app.api_route(path='/generate', methods=['POST'])
            async def generate(payload: GenerateRequest = Body(...)):
                """Generate text from a prompt."""

                parameters = payload.dict(
                    exclude_unset=True,
                    exclude_none=True,
                    exclude_defaults=True,
                    exclude={'prompt'},
                )

                parameters = _update_key(parameters)

                async for docs, error in self.streamer.stream(
                    docs=DocumentArray(
                        [
                            Document(
                                tags={'prompt': payload.prompt},
                            )
                        ]
                    ),
                    exec_endpoint='/generate',
                    parameters=parameters,
                ):
                    if error:
                        return JSONResponse(
                            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
                            content={'message': error.name},
                        )
                    else:
                        _tags = docs[0].tags.copy()
                        _tags['usage'] = {k: int(v) for k, v in _tags['usage'].items()}

                        return JSONResponse(
                            status_code=status.HTTP_200_OK,
                            content=_tags,
                        )

            @app.api_route(path='/generate_stream', methods=['POST'])
            async def generate_stream(payload: GenerateRequest = Body(...)):
                """Generate text from a prompt in streaming."""

                from fastapi import HTTPException
                from sse_starlette.sse import EventSourceResponse

                parameters = payload.dict(
                    exclude_unset=True,
                    exclude_none=True,
                    exclude_defaults=True,
                    exclude={'prompt'},
                )

                parameters = _update_key(parameters)

                async def event_generator():
                    completion_tokens = 0

                    stop_flag = False
                    while not stop_flag:
                        parameters['completion_tokens'] = completion_tokens

                        async for docs, error in self.streamer.stream(
                            docs=input_docs,
                            exec_endpoint='/generate_stream',
                            parameters=parameters,
                        ):
                            if error:
                                # TODO: find best practice to handle errors in sse
                                raise HTTPException(status_code=500, detail=error)

                            input_docs[0].tags['input_ids'] = docs[0].tags.get(
                                'output_ids'
                            )
                            input_docs[0].blob = docs[0].blob

                            stop_flag = docs[0].tags.get('choices')[0].get(
                                'finish_reason'
                            ) in [
                                'stop',
                                'length',
                            ]
                            completion_tokens += 1

                            _tags = docs[0].tags.copy()
                            for k in ['input_ids', 'output_ids', 'past_key_values']:
                                _tags.pop(k) if k in _tags else None
                            _tags['usage'] = {
                                k: int(v) for k, v in _tags['usage'].items()
                            }
                            yield {"data": json.dumps(_tags)}

                input_docs = DocumentArray(
                    [
                        Document(
                            tags={'prompt': payload.prompt},
                        )
                    ]
                )

                return EventSourceResponse(event_generator())

            return app

        jina.helper.extend_rest_interface = _extend_rest_function



"""The executor wraps the model and provides a simple way to run inference on the model."""
import pickle
from multiprocessing.pool import ThreadPool
from typing import Dict, List, Optional, Union

from docarray import DocumentArray
from jina import Executor, requests

from open_gpt.factory import create_model
from open_gpt.logs import logger


class CausualLMExecutor(Executor):
    def __init__(
        self,
        model_name_or_path: str = '',
        minibatch_size: int = 1,
        adapter_name_or_path: Optional[str] = None,
        device_map: Optional[Union[str, List[int]]] = None,
        precision: Optional[str] = None,
        num_workers: int = 4,
        max_length: int = 1024,
        **kwargs,
    ):
        super().__init__(**kwargs)

        assert (
            model_name_or_path
        ), '`model_name_or_path` must be provided to initialize the model and tokenizer.'

        self._model_name_or_path = model_name_or_path
        self._adapter_name_or_path = adapter_name_or_path
        self._minibatch_size = minibatch_size
        self._thread_pool = ThreadPool(processes=num_workers)
        self._max_length = max_length

        self.model = create_model(
            model_name_or_path,
            precision=precision,
            adapter_name_or_path=adapter_name_or_path,
            device_map=device_map,
            **kwargs,
        )

        # warmup the model to avoid the first-time slowness
        self.model.generate('Hello world!', max_new_tokens=64)

    @requests(on='/generate')
    def generate(self, docs: 'DocumentArray', parameters: Dict = {}, **kwargs):
        # TEMPORARY FIX: remove the `__results__` key from the parameters dict
        parameters.pop('__results__', None)
        max_length = int(parameters.pop('max_length', self._max_length))

        for k, v in parameters.items():
            if k in ['top_k', 'max_new_tokens', 'num_return_sequences']:
                parameters[k] = int(v)

        for d in docs:
            prompt = d.tags.get('prompt') or d.text
            if not prompt:
                continue

            d.tags.update(
                self.model.generate(prompt, max_length=max_length, **parameters)
            )

    @requests(on='/generate_stream')
    def generate_stream(self, docs: 'DocumentArray', parameters: Dict = {}, **kwargs):
        for k, v in parameters.items():
            if k in ['top_k', 'max_new_tokens', 'num_return_sequences']:
                parameters[k] = int(v)

        completion_tokens = parameters.pop('completion_tokens', 0)

        for d in docs:
            prompt = d.tags.get('prompt') or d.text
            input_ids = d.tags.get('input_ids')
            past_key_values = pickle.loads(d.blob) if len(d.blob) > 0 else None
            if not prompt:
                continue

            if input_ids is not None:
                input_ids = list(map(lambda x: int(x), input_ids))
                prompt = None

            generated_text = self.model.step_generate(
                prompt=prompt,
                input_ids=input_ids,
                past_key_values=past_key_values,
                completion_tokens=completion_tokens,
                **parameters,
            )
            resp = next(generated_text)

            d.blob = pickle.dumps(resp.pop('past_key_values'))
            d.tags.update(resp)



import logging
from typing import TYPE_CHECKING, Iterable, List, Optional, Tuple, Union, overload

import torch

if TYPE_CHECKING:
    from transformers import AutoModelForCausalLM, AutoTokenizer

from transformers.generation.logits_process import (
    LogitsProcessorList,
    RepetitionPenaltyLogitsProcessor,
    TemperatureLogitsWarper,
    TopKLogitsWarper,
    TopPLogitsWarper,
)

MIN_TEMPERATURE = 1e-5
MIN_TOP_P = 1e-8
MAX_LENGTH = 2048


def prepare_logits_processor(
    temperature: float, repetition_penalty: float, top_p: float, top_k: int
) -> LogitsProcessorList:
    processor_list = LogitsProcessorList()
    # TemperatureLogitsWarper doesn't accept 0.0
    # 1.0 makes it a no-op so we skip two cases.
    if MIN_TEMPERATURE <= temperature < 1:
        processor_list.append(TemperatureLogitsWarper(temperature))
    if repetition_penalty > 1:
        processor_list.append(RepetitionPenaltyLogitsProcessor(repetition_penalty))
    if MIN_TOP_P <= top_p < 1:
        processor_list.append(TopPLogitsWarper(top_p))
    if top_k > 0:
        processor_list.append(TopKLogitsWarper(top_k))
    return processor_list


def partial_stop(output, stop_str):
    for i in range(0, min(len(output), len(stop_str))):
        if stop_str.startswith(output[-i:]):
            return True
    return False


class GenerationMixin:
    """Mixin for generation methods."""

    model: 'AutoModelForCausalLM'
    tokenizer: 'AutoTokenizer'

    @torch.inference_mode()
    def step_generate(
        self,
        prompt: Optional[str] = None,
        input_ids: Optional[List[int]] = None,
        max_new_tokens: Optional[int] = None,
        completion_tokens: int = 0,
        temperature: float = 1.0,
        top_k: int = 1,
        top_p: float = 0.9,
        repetition_penalty: float = 1.0,
        max_length: int = MAX_LENGTH,
        echo: bool = False,
        stream_interval: int = 1,
        stop_str: Optional[str] = None,
        stop_token_ids: List[int] = [],
        past_key_values: Optional[Iterable[torch.Tensor]] = None,
        **kwargs,
    ):
        """Generate tokens in a streaming fashion. This method is a modified version of `fastchat.server.inference.generate_stream`.

        :param prompt: The prompt is the context that the model will use to generate the response.
        :param input_ids: The input ids to use for generation.
        :param max_new_tokens: The maximum number of tokens to generate. If None, the model will generate until it predicts a stop token.
        :param completion_tokens: The number of tokens that has been generated before.
        :param temperature: The temperature to use when sampling from the logits.
        :param top_k: The number of highest probability vocabulary tokens to keep for top-k-filtering.
        :param top_p: The cumulative probability of parameter highest probability vocabulary tokens to keep for nucleus sampling.
        :param repetition_penalty: The parameter for repetition penalty. 1.0 means no penalty. See `this paper <https://arxiv.org/pdf/1909.05858.pdf>`__ for more details.
        :param max_length: Maximum length of the sequence which includes prompt and generated text. If the context is longer than this, it will be truncated.
        :param echo: If True, the prompt will be included in the generated response.
        :param stream_interval: The number of tokens to generate before returning the generated tokens.
        :param stop_str: If not None, the model will stop generating when the generated tokens end with this string.
        :param stop_token_ids: A list of token ids that will cause the model to stop generating.
        :param past_key_values: A list of past key values to use for generation. If None, the model will generate from scratch.
        :param kwargs: Additional keyword arguments to pass to the model.
        :return: A dictionary contains generated text, output ids, past_key_values, finish reason and usage information.
                usage information: {'completion_tokens': int, how many tokens have been generated.
                                    'input_length': int, the length of input ids or prompt.
                                    'prompt_tokens': int, the length of past_key_values passed to model.forward() when generating this token.
                                    'total_tokens': int, completed_tokens + input_tokens.
                                }
        """

        def _get_finish_reason(step, completion_tokens, max_new_tokens):
            if step + 1 + completion_tokens == max_new_tokens:
                return "length"
            elif stopped:
                return "stop"
            else:
                return None

        if prompt is None and input_ids is None:
            raise ValueError("Either prompt or input_ids must be provided.")
        if prompt and input_ids:
            raise ValueError("Only one of prompt or input_ids can be provided.")

        if input_ids is None:
            input_ids = self.tokenizer(prompt, return_tensors="pt")["input_ids"].to(
                self._device
            )
            len_prompt = len(input_ids[0])
        else:
            assert isinstance(input_ids[0], int), (
                f"input_ids must be list of int, " f"got list of {type(input_ids[0])}"
            )
            len_prompt = None
            input_ids = torch.Tensor([input_ids]).to(dtype=int).to(self._device)

        input_length = len(input_ids[0])

        logits_processor = prepare_logits_processor(
            temperature,
            repetition_penalty,
            top_p,
            top_k,
        )

        stop_token_ids.append(self.tokenizer.eos_token_id)

        output_ids = input_ids.tolist()[0]

        # TODO: figure out what 8 means here
        max_src_len = max_length - max_new_tokens - 8
        if input_length > max_src_len:
            input_ids = input_ids[:, -max_src_len:]
            input_length = max_src_len

        next_token = None

        for step in range(max_new_tokens):
            prompt_tokens = past_key_values[0][0].shape[2] if past_key_values else 0
            if step == 0:
                outputs = self.model(
                    input_ids, use_cache=True, past_key_values=past_key_values
                )
                logits = outputs.logits
                past_key_values = outputs.past_key_values
            else:
                outputs = self.model(
                    input_ids=torch.as_tensor([[next_token]], device=self._device),
                    use_cache=True,
                    past_key_values=past_key_values,
                )
                logits = outputs.logits
                past_key_values = outputs.past_key_values

            if logits_processor:
                if repetition_penalty > 1.0:
                    tmp_output_ids = torch.as_tensor([output_ids], device=logits.device)
                else:
                    tmp_output_ids = None
                last_token_logits = logits_processor(tmp_output_ids, logits[:, -1, :])[
                    0
                ]
            else:
                last_token_logits = logits[0, -1, :]

            if last_token_logits.device.type == "mps":
                # Switch to CPU by avoiding some bugs in mps backend.
                last_token_logits = last_token_logits.float().to("cpu")

            if temperature < MIN_TEMPERATURE or top_p < MIN_TOP_P:  # greedy
                next_token = int(torch.argmax(last_token_logits))
            else:
                probs = torch.softmax(last_token_logits, dim=-1)
                next_token = int(torch.multinomial(probs, num_samples=1))

            output_ids.append(next_token)
            stopped = next_token in stop_token_ids

            if (
                step % stream_interval == 0
                or step + completion_tokens == max_new_tokens - 1
                or stopped
            ):
                if echo and len_prompt is not None:
                    # TODO: this is a HOTFIX to keep the same behavior as setting echo=False
                    # tmp_output_ids = output_ids
                    # rfind_start = len_prompt
                    tmp_output_ids = output_ids[input_length:]
                    rfind_start = 0
                else:
                    if echo and len_prompt is None:
                        logging.warning(
                            f"echo is set to True but prompt is not provided. "
                            f"Back to non-echo mode."
                        )
                    tmp_output_ids = output_ids[input_length:]
                    rfind_start = 0

                output = self.tokenizer.decode(
                    tmp_output_ids,
                    skip_special_tokens=True,
                    spaces_between_special_tokens=False,
                )

                partially_stopped = False
                if stop_str:
                    if isinstance(stop_str, str):
                        pos = output.rfind(stop_str, rfind_start)
                        if pos != -1:
                            output = output[:pos]
                            stopped = True
                        else:
                            partially_stopped = partial_stop(output, stop_str)
                    elif isinstance(stop_str, Iterable):
                        for each_stop in stop_str:
                            pos = output.rfind(each_stop, rfind_start)
                            if pos != -1:
                                output = output[:pos]
                                stopped = True
                                break
                            else:
                                partially_stopped = partial_stop(output, each_stop)
                                if partially_stopped:
                                    break
                    else:
                        raise ValueError("Invalid stop field type.")

                # prevent yielding partial stop sequence
                if not partially_stopped:
                    yield {
                        "choices": [
                            {
                                "index": step,
                                "text": output,
                                "finish_reason": _get_finish_reason(
                                    step, completion_tokens, max_new_tokens
                                ),
                            },
                        ],
                        "prompt": prompt,
                        "output_ids": tmp_output_ids,
                        "past_key_values": past_key_values,
                        "usage": {
                            "prompt_tokens": prompt_tokens,
                            "input_length": input_length,
                            "completion_tokens": completion_tokens + step + 1,
                            "total_tokens": prompt_tokens + 1,
                        },
                    }

            if stopped:
                break

        yield {
            "choices": [
                {
                    "index": step,
                    "text": output,
                    "finish_reason": _get_finish_reason(
                        step, completion_tokens, max_new_tokens
                    ),
                },
            ],
            "prompt": prompt,
            "output_ids": tmp_output_ids,
            "past_key_values": past_key_values,
            "usage": {
                "prompt_tokens": prompt_tokens,
                "input_length": input_length,
                "completion_tokens": completion_tokens + step + 1,
                "total_tokens": prompt_tokens + 1,
            },
        }

    @overload
    def generate(self, prompt: str, inplace_images: List = [], **kwargs):
        ...

    @overload
    def generate(
        self,
        prompt: str,
        max_new_tokens: Optional[int] = None,
        num_beams: int = 1,
        do_sample: bool = False,
        temperature: float = 1.0,
        top_k: int = 1,
        top_p: float = 0.9,
        repetition_penalty: float = 1.0,
        length_penalty: float = 1.0,
        no_repeat_ngram_size: int = 0,
        echo: bool = False,
        **kwargs,
    ):
        """Generate text from the given prompt.

        :param prompt: The prompt input text.
        :param max_new_tokens: The maximum number of tokens to generate, not including the prompt.
        :param num_beams: Number of beams for beam search. 1 means no beam search.
        :param do_sample: Whether to use sampling instead of greedy decoding.
        :param temperature: The temperature to use for sampling. Only relevant if do_sample is True. Higher means more stochastic.
        :param top_k: The number of highest probability vocabulary tokens to keep for top-k-filtering. Only relevant if do_sample is True.
        :param top_p: The cumulative probability of parameter highest probability vocabulary tokens to keep for nucleus sampling. Only relevant if do_sample is True.
        :param repetition_penalty: The parameter for repetition penalty. 1.0 means no penalty.
        :param length_penalty: Exponential penalty to the length that is used with beam-based generation.
                It is applied as an exponent to the sequence length, which in turn is used to divide the score of the sequence.
                Since the score is the log likelihood of the sequence (i.e. negative), length_penalty > 0.0 promotes longer sequences,
                while length_penalty < 0.0 encourages shorter sequences.
        :param no_repeat_ngram_size: If set to int > 0, all ngrams of that size can only occur once.
        :param echo: Whether to echo the prompt in the generated text.
        """
        ...

    def generate(self, prompt: str, max_length: int = MAX_LENGTH, **kwargs):
        inputs = self.tokenizer(
            prompt,
            return_tensors="pt",
        )
        inputs.pop("token_type_ids", None)
        inputs = inputs.to(self._device)

        input_length = inputs["input_ids"].shape[-1]

        # validate kwargs
        if 'stop_str' in kwargs:
            kwargs.pop('stop_str')
            logging.warning(f"stop_str is not supported in generate mode.")

        # overwrite default values with kwargs
        clean_up_tokenization_spaces = kwargs.pop('clean_up_tokenization_spaces', True)
        skip_special_tokens = kwargs.pop("skip_special_tokens", True)
        echo = kwargs.pop("echo", False)

        max_length = kwargs.pop("max_length", max_length)
        max_new_tokens = kwargs.pop("max_new_tokens", max_length - input_length - 1)

        if max_new_tokens + input_length >= max_length:
            raise ValueError(
                f"max_new_tokens + input_length ({max_new_tokens + input_length}) must be less than max_length ({max_length})"
            )

        kwargs.update(
            {
                "max_length": max_length,
                "max_new_tokens": max_new_tokens,
            }
        )

        with torch.inference_mode():
            outputs = self.model.generate(**inputs, **kwargs)[0].tolist()
            text = self.tokenizer.decode(
                outputs if echo else outputs[input_length:],
                clean_up_tokenization_spaces=clean_up_tokenization_spaces,
                skip_special_tokens=skip_special_tokens,
            )

            resp = {
                "choices": [
                    {"index": 0, "text": text, "finish_reason": "length"},
                ],
                "prompt": prompt,
                "usage": {
                    "prompt_tokens": 0,
                    "input_length": input_length,
                    "completion_tokens": len(outputs),
                    "total_tokens": input_length + len(outputs),
                },
            }

            return resp
